/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.kafka.common.security.ldap.internals;

import static org.junit.jupiter.api.Assertions.assertThrows;

import org.apache.directory.api.ldap.model.entry.DefaultEntry;
import org.apache.directory.api.ldap.model.exception.LdapException;
import org.apache.directory.api.ldap.model.ldif.LdifEntry;
import org.apache.directory.api.ldap.model.ldif.LdifReader;
import org.apache.directory.api.util.FileUtils;
import org.apache.directory.server.core.factory.DefaultDirectoryServiceFactory;
import org.apache.directory.server.core.integ.AbstractLdapTestUnit;
import org.apache.directory.server.i18n.I18n;
import org.apache.directory.server.ldap.LdapServer;
import org.apache.directory.server.protocol.shared.transport.TcpTransport;
import org.apache.kafka.common.errors.SaslAuthenticationException;
import org.apache.kafka.common.security.JaasContext;
import org.apache.kafka.common.security.authenticator.TestJaasConfig;
import org.apache.kafka.common.security.plain.PlainLoginModule;
import org.apache.kafka.common.security.plain.internals.PlainSaslServer;
import org.apache.kafka.common.security.plain.internals.PlainServerCallbackHandler;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.net.ServerSocket;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class LdapAuthenticationTest extends AbstractLdapTestUnit {

    private PlainSaslServer saslServer;

    private static final String USER_ID = "kafkaUser";
    private static final String USER_PASSWORD = "ldappassword";
    private static final String BASE_PARTITION_NAME = "mydomain";
    private static final String BASE_DOMAIN = "org";
    private static final String BASE_STRUCTURE = "dc=" + BASE_PARTITION_NAME + ",dc=" + BASE_DOMAIN;
    private static final String LDIF = "dn: uid=kafkaUser,ou=users,ou=system\n" +
        "objectClass: uidObject\n" +
        "objectClass: person\n" +
        "objectClass: top\n" +
        "uid: kafkauUser\n" +
        "cn: Kafka User\n" +
        "sn: Kafka\n" +
        "userPassword: ldappassword";

    @BeforeAll
    public static void init() throws Exception {
        ServerSocket ss = new ServerSocket(0);
        int port = ss.getLocalPort();
        ss.close();

        DefaultDirectoryServiceFactory serviceFactory = new DefaultDirectoryServiceFactory();
        serviceFactory.init("LdapAuthenticationTest_DS");
        setService(serviceFactory.getDirectoryService());
        try (LdifReader reader = new LdifReader()) {
            List<LdifEntry> entries = reader.parseLdif(LDIF);
            for (LdifEntry entry : entries) {
                if (entry.isChangeAdd() || entry.isLdifContent()) {
                    getService().getAdminSession().add(new DefaultEntry(getService().getSchemaManager(), entry.getEntry()));
                } else if (entry.isChangeModify()) {
                    getService().getAdminSession().modify(entry.getDn(), entry.getModifications());
                } else {
                    throw new LdapException(I18n.err(I18n.ERR_117, entry.getChangeType()));
                }
            }
        }

        getService().startup();
        setLdapServer(new LdapServer());
        getLdapServer().setDirectoryService(service);
        getLdapServer().setTransports(new TcpTransport(port));
        getLdapServer().start();
    }

    @BeforeEach
    public void setUp() {
        TestJaasConfig jaasConfig = new TestJaasConfig();
        Map<String, Object> options = new HashMap<>();
        options.put("ldap_url", "ldap://localhost:" + getLdapServer().getPort());
        options.put("user_dn_template", "uid={0},ou=users,ou=system");
        jaasConfig.addEntry("jaasContext", PlainLoginModule.class.getName(), options);
        JaasContext jaasContext = new JaasContext("jaasContext", JaasContext.Type.SERVER, jaasConfig, null);
        PlainServerCallbackHandler callbackHandler = new LdapPlainServerCallbackHandler();
        callbackHandler.configure(null, "PLAIN", jaasContext.configurationEntries());
        saslServer = new PlainSaslServer(callbackHandler);
    }

    @AfterAll
    public static void destroy() throws Exception {
        File instanceDirectory = getService().getInstanceLayout().getInstanceDirectory();
        getLdapServer().stop();
        getService().shutdown();
        setLdapServer(null);
        setService(null);
        FileUtils.deleteDirectory(instanceDirectory);
    }

    @Test
    public void testSuccessfulAuthentication() {
        saslServer.evaluateResponse(saslMessage(USER_PASSWORD));
    }

    @Test
    public void testFailingAuthentication() {
        assertThrows(
            SaslAuthenticationException.class,
            () -> saslServer.evaluateResponse(saslMessage("incorrect_password"))
        );
    }

    private byte[] saslMessage(String password) {
        String nul = "\u0000";
        String message = String.format("%s%s%s%s%s", LdapAuthenticationTest.USER_ID, nul, LdapAuthenticationTest.USER_ID, nul, password);
        return message.getBytes(StandardCharsets.UTF_8);
    }
}
